# Standard Library
from typing import TYPE_CHECKING

# External Dependencies
from dynaconf import Dynaconf
from pydantic import BaseSettings, Extra

# Shakah Library
from shakah.dysettings import settings as dyset


class DynaEnabledSettings(BaseSettings):
    class Config:
        case_sensitive = False
        env_file = ".env"
        extra = Extra.allow
        smart_union = True

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                dynaconf_source,
                init_settings,
                env_settings,
                file_secret_settings,
            )


def dynaconf_source(settings: BaseSettings) -> Dynaconf:
    with dyset.fresh():
        return dyset



# Standard Library
import uuid
from typing import Optional

# External Dependencies
from loguru import logger
from pydantic import Extra
from pydantic.networks import EmailStr
from pydantic.types import conint, constr
from sqlmodel.main import Relationship, SQLModel

# Shakah Library
from shakah import Field, ShakahModel


class User(ShakahModel, table=True):
    email: EmailStr
    email_verified: bool
    name: str
    given_name: str
    family_name: str
    preferred_username: str
    nickname: str
    token_id: Optional[uuid.UUID] = Field(default=None, foreign_key="oauthtoken.id")
    token: Optional["OAuthToken"] = Relationship(back_populates="user")


class OAuthToken(ShakahModel, table=True):
    name: str
    token_type: str
    access_token: str
    refresh_token: str
    expires_at: conint(gt=0)
    user: Optional[User] = Relationship(back_populates="token")

    class Config:
        extra = Extra.ignore

    def to_token(self):
        return dict(
            access_token=self.access_token,
            token_type=self.token_type,
            refresh_token=self.refresh_token,
            expires_at=self.expires_at,
        )



# Standard Library
import abc
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, List, Optional, Protocol, Type, Union

# External Dependencies
from pydantic import AnyUrl, BaseModel
from pydantic.networks import PostgresDsn, RedisDsn

if TYPE_CHECKING:
    from pydantic.networks import Parts

# Shakah Library
from shakah import logger

# A Url Protocol for SQLite


class EnvType(str, Enum):
    LOCAL = ""
    DEV = "developer"
    PROD = "admin"



# External Dependencies

# External Dependencies
from loguru import logger
from sqlmodel import Field

# Shakah Library
from shakah._core.definitions import ShakahModel

# from ._core.definitions import ShakahModel
# from ._core.sourcing import DynaEnabledSettings

# from .config import settings



# Standard Library
from typing import Any, Dict, Optional, Union

# External Dependencies
from loguru import logger
from pydantic import AnyUrl, PostgresDsn, validator

# Shakah Library
from shakah._core.sourcing import DynaEnabledSettings
from shakah.typings import SqliteUrl
from shakah.utils.knoting import build_connection_string


class Settings(DynaEnabledSettings):

    PORT: int
    NAME: str
    SCHEME: str
    HOST: str
    DATABASE: str
    USER: str
    PASSWORD: str

    DATABASE_URI: Union[SqliteUrl, PostgresDsn, AnyUrl, str, None] = None

    @validator("DATABASE_URI", pre=True)
    def generate_url(cls, v: Optional[str], values: Dict[str, Any]) -> Any:
        if isinstance(v, str):
            return v
        lower_values = {k.lower(): v for k, v in values.items()}
        scheme = lower_values.pop("scheme")
        conn_str = build_connection_string(scheme, lower_values)
        logger.warning(conn_str)
        return conn_str


settings = Settings()



# Standard Library
import inspect
from typing import Any, Callable, Dict, Optional, Tuple, Type, TypeVar, Union

# External Dependencies
import wrapt
from fastcore.basics import annotations
from loguru import logger
from sqlalchemy import or_
from sqlalchemy.future.engine import Engine
from sqlmodel import Field, Session, SQLModel, select
from sqlmodel.engine.result import Result, ScalarResult
from sqlmodel.main import FieldInfo, SQLModelMetaclass
from sqlmodel.sql.expression import Select, SelectOfScalar

# Shakah Library
from shakah import ShakahModel
from shakah.conn import get_engine, push_event, subscribe
from shakah.utils import classproperty
from shakah.utils.decorators.classes import ClassPropertyMetaClass


class ExampleModel(ShakahModel):

    post_text: str = Field(None, index=True)

    class Config:
        table = True
        extend_existing = True


def main():

    SQLModel.metadata.create_all(ShakahModel.engine_)
    built = ExampleModel(post_text="Hello World")
    logger.info(built)
    built.save()
    logger.info(built)
    # found.first()
    # debug(found.one())
    push_event("engine_update", "metadata")


if __name__ == "__main__":
    main()



# Standard Library
import abc
import inspect
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

# External Dependencies
from decorator import decorator
from pydantic import BaseModel
from pydantic.main import ModelMetaclass
from sqlalchemy import or_
from sqlalchemy.future.engine import Engine
from sqlmodel import Field, Session, SQLModel, create_engine, select
from sqlmodel.main import FieldInfo, SQLModelMetaclass, resolve_annotations

# Shakah Library
from shakah import logger, typings
from shakah.config import settings
from shakah.utils import classproperty
from shakah.utils.decorators.classes import (
    ClassPropertyDescriptor,
    ClassPropertyMetaClass,
)

logger.disable("shakah.conn")


EXISTING_ENGINE = None
SUBSCRIBERS: Dict[str, List[Type[typings.SessionKeep]]] = {}
SUBSCRIBER_SET = set()


def sublen() -> int:
    return len(SUBSCRIBER_SET)


def subscribe(event_type: str, cls_: Type[Any]):
    setlen = sublen()
    if hasattr(cls_, "mname"):
        SUBSCRIBER_SET.add(cls_.mname)
        if sublen() == setlen:
            return
    logger.warning(f"Subscribing - {cls_.__name__}")
    global SUBSCRIBERS
    if not event_type in SUBSCRIBERS:
        SUBSCRIBERS[event_type] = []
    SUBSCRIBERS[event_type].append(cls_)


def push_event(event_type: str, data: Any):
    global SUBSCRIBERS
    if not event_type in SUBSCRIBERS:
        SUBSCRIBERS[event_type] = []
        return
    event_subscribers = SUBSCRIBERS.get(event_type, [])
    for subscriber in event_subscribers:
        logger.success(subscriber)
        subscriber().update(data)  # type: ignore


def set_engine(engine_: Optional[Engine]):
    global EXISTING_ENGINE
    EXISTING_ENGINE = engine_ or create_engine(str(settings.DATABASE_URI))
    push_event("engine_update", "metadata")


def get_engine() -> Engine:

    # Update the engine every time you change the parameters. Register to model to change engine.
    global EXISTING_ENGINE
    if EXISTING_ENGINE is None:
        EXISTING_ENGINE = create_engine(str(settings.DATABASE_URI))
    return EXISTING_ENGINE



# Standard Library
from typing import TYPE_CHECKING, Any, Optional, Protocol, Type, Union

# External Dependencies
from pydantic import AnyUrl, BaseModel
from pydantic.networks import PostgresDsn as _PostgresDsn
from pydantic.networks import RedisDsn

if TYPE_CHECKING:
    from pydantic.networks import Parts

# Shakah Library
from shakah import logger

# A Url Protocol for SQLite


class SqliteUrl(AnyUrl):
    allowed_schemes = {"sqlite"}
    host_required = False
    hidden_parts = {"port"}

    @staticmethod
    def get_default_parts(parts: "Parts") -> "Parts":
        {}
        return {
            "domain": "",
            "port": "" if parts["port"] in ["80", None] else parts["port"],
            "path": "/:memory:",
        }


# A Postgresql AnyURL with default parts for localhost
class PostgresDsn(_PostgresDsn):
    @staticmethod
    def get_default_parts(parts: "Parts") -> "Parts":
        return {
            "domain": "localhost" if parts["domain"] in ["", None] else parts["domain"],
            "port": "5432",
            "path": "app" if parts["path"] in ["", None] else parts["path"],
        }


class URIModel(BaseModel):
    url_type: Optional[Type[Union[PostgresDsn, SqliteUrl, RedisDsn, AnyUrl]]] = None
    port: Optional[int] = None
    name: Optional[str] = None
    scheme: Optional[str] = None
    host: Optional[str] = None
    database: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None

    def build(self) -> str:
        parts: Parts = self.dict(exclude={"url_type"})  # type: ignore
        parts["domain"] = parts.pop("host")  # type: ignore
        parts["path"] = parts.pop("database")  # type: ignore
        parts["port"] = str(parts.pop("port"))  # type: ignore
        outcome = self.url_type.apply_default_parts(parts)  # type: ignore
        if self.url_type is not None:
            built = self.url_type.build(
                scheme=str(self.scheme),
                user=str(outcome["user"]),
                password=str(outcome["password"]),
                host=str(outcome["domain"]),
                path=f"/{outcome['path']}",
                port=str(outcome["port"]),
            )
            return str(built)
        raise AttributeError("The URL Scheme Is Not Set")
        # self.url_type.build()


def select_uri(scheme: str) -> URIModel:
    return {
        "sqlite": SqliteUrl,
        "postgresql": PostgresDsn,
    }[scheme]


def build_connection_string(scheme: str, data: dict) -> str:
    uri_model = URIModel(url_type=select_uri(scheme), scheme=scheme, **data)
    logger.info(uri_model)
    built = uri_model.build()
    return built


# -----------------------------------------------------------------------------
#                               Protocols
# -----------------------------------------------------------------------------
class SessionKeep(Protocol):
    def instance(self):
        pass

    # @abc.abstractmethod
    def update(self, data: Any):
        raise NotImplementedError



# Standard Library
import inspect
import uuid
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

# External Dependencies
import wrapt
from fastcore.basics import annotations
from loguru import logger
from sqlalchemy import func, or_
from sqlalchemy.engine import Engine
from sqlmodel import Field, Session, SQLModel, select
from sqlmodel.engine.result import Result, ScalarResult
from sqlmodel.main import FieldInfo, SQLModelMetaclass
from sqlmodel.sql.expression import Select, SelectOfScalar

# Shakah Library
from shakah.conn import get_engine, subscribe
from shakah.errors import RecordExistError
from shakah.utils import classproperty
from shakah.utils.decorators.classes import ClassPropertyMetaClass

logger.disable("shakah._core.definitions")

_T = TypeVar("_T")


def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: Tuple[Union[type, Callable[..., Any]], ...] = (()),
) -> Callable[[_T], _T]:
    return lambda a: a


@wrapt.decorator
def autoselect(wrapped, instance, args, kwargs):
    # logger instance and arg variables
    return instance.db.exec(wrapped(*args, **kwargs))


@wrapt.decorator
def unwrapfirst(wrapped, instance, args, kwargs):
    # logger instance and arg variables
    res = wrapped(*args, **kwargs)
    if not res:
        return res
    return res.first()


def criterion(x):
    if not inspect.isroutine(x):
        return False
    name = x.__name__

    if name.startswith("__") or name.startswith("_"):
        return False

    if name in ["parse_raw", "schema", "schema_json", "parse_file", "construct"]:
        return False
    annos = {}
    try:
        annos = annotations(x)
    except Exception as e:
        logger.error(f"Can't extract annotations for {name}")
        return False
    if not annos:
        return False
    returns = annos.get("return", None)
    if not returns:
        return False
    if returns not in [SelectOfScalar, Select]:
        return False
    return True


@__dataclass_transform__(kw_only_default=True, field_descriptors=(Field, FieldInfo))
class ShakahModelBuiltinMetaclass(SQLModelMetaclass, ClassPropertyMetaClass):
    __engine__: Engine
    __session__: Session
    __updb__: bool = False
    __instance__: Any = None
    __metaname__: str = "ShakahModelMetaclass"

    # create unique id for each model

    def __new__(
        cls,
        name: str,
        bases: Tuple[Type[Any], ...],
        class_dict: Dict[str, Any],
        **kwargs: Any,
    ) -> Any:
        model = super().__new__(cls, name, bases, class_dict, **kwargs)
        model.update_meta()
        subscribe("engine_update", model)
        # create singleton instance
        if not hasattr(model, "__instance__"):
            model.__instance__ = model()

        return model

    @classmethod
    def update_engine(cls):

        setattr(cls, "__engine__", get_engine())
        setattr(cls, "__updb__", True)
        return True

    @classmethod
    def update_session(cls):
        if not getattr(cls, "__session__", None) or cls.__updb__ == True:

            setattr(cls, "__session__", Session(cls.__engine__))
            setattr(cls, "__updb__", False)

    @classproperty
    def engine_(cls):

        return cls.__engine__

    @engine_.setter  # type: ignore
    def engine_(cls, value):
        cls.__engine__ = value

    @classproperty
    def db_(cls):

        return cls.__session__

    @classmethod
    def update_meta(cls):
        cls.update_engine()
        cls.update_session()

    def getty(cls, name: str, default=None):
        return getattr(cls, name, default)

    # putty set attr
    def putty(cls, name: str, value):
        setattr(cls, name, value)

    # Get instance with classproperty

    @classproperty
    def instance_(cls):
        return cls.__instance__

    @classproperty
    def mname(cls):
        return cls.__metaname__

    def update(self, data: str):
        if data == "metadata":
            logger.info("Updating engine & session")
            self.md.update_meta()


class ShakahProps(SQLModel, metaclass=ShakahModelBuiltinMetaclass):
    __table_args__ = {"extend_existing": True}

    def __init__(self_, **data: Any) -> None:
        super().__init__(**data)

    @property
    def md(s):
        return type(s)

    @property
    def select(self):
        return select(self.md)

    @property
    def engine(s):
        return s.md.engine_

    @property
    def db(s) -> Session:
        return s.md.db_

    @property
    def active(self):
        return self.dict(exclude_none=True, exclude_unset=True, exclude_defaults=True)

    @property
    def is_active(self) -> bool:
        return bool(self.active)

    def binops(self, exclude_id: bool = True) -> List[Any]:
        def compare(key, value):
            return getattr(self.md, key) == value

        bins = []
        for k, v in self.active.items():
            if exclude_id and k == "id":
                continue
            bins.append(compare(k, v))
        return bins

    @autoselect
    def find_(self, *args, exclude_id: bool = True, **kwargs) -> SelectOfScalar:
        if not self.is_active:
            raise ValueError("No data was inputted into the system.")
        return self.select.where(or_(*self.binops(exclude_id=exclude_id)))


class ShakahModel(ShakahProps, SQLModel):
    id: Optional[uuid.UUID] = Field(primary_key=True, default_factory=uuid.uuid4)

    @unwrapfirst
    @autoselect
    def count(self):
        return select(func.count()).select_from(self.md)

    @unwrapfirst
    @autoselect
    def count_match(self, exclude_id: bool = False):
        if not self.is_active:
            raise ValueError("No data was inputted into the system.")
        return (
            select(func.count())
            .select_from(self.md)
            .where(or_(*self.binops(exclude_id=exclude_id)))
        )

    def sync(self, *args, **kwargs) -> bool:
        # vdict = s.active
        update_data = self.dict(exclude_unset=True)
        first_record = self.find().first()
        if not first_record:
            return False

        first_record.update(update_data)
        self = self.copy(update=first_record)

        return True

    def save(self):

        self.db.add(self)
        self.db.commit()
        self.db.refresh(self)

    def delete(self, *args, **kwargs):
        record = self.find().first(*args, **kwargs)
        if not record:
            raise RecordExistError("Record not found.")
        # with self.db as db:
        self.db.delete(record)

    def shift(self):
        first_record = self.find().first()
        if not first_record:
            return False

    def find(self, *args, **kwargs) -> Union[ScalarResult, Result]:
        found: ScalarResult = self.find_(*args, **kwargs)  # type: ignore
        return found

    def find_by_id(self, id: str):
        return self.md(id=id).find_(exclude_id=False).first()



# Standard Library
from typing import TYPE_CHECKING, Optional, Type, Union

# External Dependencies
from pydantic import AnyUrl, BaseModel
from pydantic.networks import RedisDsn

if TYPE_CHECKING:
    from pydantic.networks import Parts

# Shakah Library
from shakah.typings import PostgresDsn, SqliteUrl

# A Url Protocol for SQLite


class URIModel(BaseModel):
    url_type: Optional[Type[Union[PostgresDsn, SqliteUrl, RedisDsn, AnyUrl]]] = None
    port: Optional[int] = None
    name: Optional[str] = None
    scheme: Optional[str] = None
    host: Optional[str] = None
    database: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None

    def build(self) -> str:
        parts: Parts = self.dict(exclude={"url_type"})  # type: ignore
        parts["domain"] = parts.pop("host")  # type: ignore
        parts["path"] = parts.pop("database")  # type: ignore
        parts["port"] = str(parts.pop("port"))  # type: ignore
        outcome = self.url_type.apply_default_parts(parts)  # type: ignore
        if self.url_type is not None:
            built = self.url_type.build(
                scheme=str(self.scheme),
                user=str(outcome["user"]),
                password=str(outcome["password"]),
                host=str(outcome["domain"]),
                path=f"/{outcome['path']}",
                port=str(outcome["port"]),
            )
            return str(built)
        raise AttributeError("The URL Scheme Is Not Set")
        # self.url_type.build()



# Standard Library
from typing import TYPE_CHECKING, Any, Protocol

if TYPE_CHECKING:
    pass

# Shakah Library
from shakah import logger
from shakah.gears.structs.vehicles import URIModel
from shakah.typings import PostgresDsn, SqliteUrl

# A Url Protocol for SQLite


def select_uri(scheme: str) -> URIModel:
    return {
        "sqlite": SqliteUrl,
        "postgresql": PostgresDsn,
    }[scheme]


def build_connection_string(scheme: str, data: dict) -> str:
    uri_model = URIModel(url_type=select_uri(scheme), scheme=scheme, **data)
    built = uri_model.build()
    return built


# -----------------------------------------------------------------------------
#                               Protocols
# -----------------------------------------------------------------------------
class SessionKeep(Protocol):
    def instance(self):
        pass

    # @abc.abstractmethod
    def update(self, data: Any):
        raise NotImplementedError



class ClassPropertyDescriptor(object):

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self

def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)

class ClassPropertyMetaClass(type):
    def __setattr__(self, key, value):
        if key in self.__dict__:
            obj = self.__dict__.get(key)
        if obj and type(obj) is ClassPropertyDescriptor:
            return obj.__set__(self, value)

        return super(ClassPropertyMetaClass, self).__setattr__(key, value)


# Create class that decorates all functions with return type in annotation
# class ShakahModelBuiltinMetaclass(type):
    
    


# External Dependencies
from dynaconf import Dynaconf
from dynaconf.loaders import redis_loader
from loguru import logger

# from dynaconf.loaders
settings = Dynaconf(
    envvar_prefix="DYNACONF",
    settings_files=[
        "settings.toml",
    ],
    environments=True,
    lowercase_read=True,
    redis_enabled=True,
)
# logger.info(settings.randovalue)


redis_loader.write(
    settings,
    dict(
        SECRET="redis_works",
        POSTGRES_USER="postgres",
        POSTGRES_PASSWORD="postgres",
        POSTGRES_SERVER="localhost",
        POSTGRES_DB="app",
        POSTGRES_DB_TWO="app2",
        POSTGRES_DB_THREE="app3",
        **{"dynaconf_merge": True}
    ),
)

# with settings.using_env("development"):
#     redis_loader.write(settings, {"SECRET": "redis_works_in_dev"})
# `envvar_prefix` = export envvars with `export DYNACONF_FOO=bar`.
# `settings_files` = Load these files in the order.



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['shakah',
 'shakah._core',
 'shakah.gears',
 'shakah.gears.structs',
 'shakah.models',
 'shakah.utils',
 'shakah.utils.decorators']

package_data = \
{'': ['*']}

install_requires = \
['decorator>=5.1.1,<6.0.0',
 'devtools>=0.8.0,<0.9.0',
 'dynaconf[redis,vault]>=3.1.7,<4.0.0',
 'fastcore>=1.3.27,<2.0.0',
 'import-linter>=1.2.6,<2.0.0',
 'loguru>=0.5.3,<0.6.0',
 'psycopg2>=2.9.3,<3.0.0',
 'pydantic[dotenv,email]>=1.9.0,<2.0.0',
 'sqlmodel>=0.0.6,<0.0.7']

setup_kwargs = {
    'name': 'shakah',
    'version': '1.1.0',
    'description': '',
    'long_description': '',
    'author': 'kivo360',
    'author_email': 'kivo360@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
